;Proof of concept of using the TAS instruction (opcode $9b) for changing sprite image pointers automatically
;as a function of some loop counter. For a complete description of this opcode, read the document written
;by Groepaz (http://csdb.dk/release/?id=143981).
;
;I use TAS as follows. I switch banks and set screen memory to $7800. Then the sprite image pointers are at $7bf8-$7bff.
;The code of interest is:
; LAX Counter 
; TAS $7bf8,y
;LAX sets A  and X to the countervalue.
;TAS performs A and X, and makes the Stack Pointer (SP) equal to that value (A and X are not modified).
;TAS then ANDs SP with $7c, and stores that in $7bf8,y, selecting a certain sprite image that depends on the countervalue. Since $7c has bits 2-6 set, TAS can select 32 different sprite images.
;Here a countervalue is used, but the registers could also be made to depend in some complicated way on the sprite coordinates. For example, even on a Z coordinate if you want to do 3D stuff.
;I also use the fact that SP has been modified, by pulling a number from the stack that represents sprite color and the color put on the screen underneath the sprite. The random numbers are refreshed from time to time.


*=$801

        BYTE    $0B, $08, $0A, $00, $9E, $32, $30, $36, $34, $00, $00, $00

rnd1 = $02      ; Low byte Random Number Generator.
rnd2 = $03      ; High byte RNG.
temp = $04      ; Temporary storage.
xmsb = $05      ; to keep track of msb x-position.
movementCounter = $06 ; Sprites are moving outwards as long as this counter > 0.
numMoves = $81  ; This many updates in sprite positions.
numReps = 10    ;Refresh random numbers after a number of repetitions 
repCounter = $07 ;How often the sprites are launched before refreshing random numbers on the stack.
safeSP = $f8    ;Some safe high stack pointer value.
repeat = $08    ;To count down twice the movementCounter before refreshing random numbers (This makes the sprites spiral back in).
musicCounter = $09      ; 2 bytes. when the counter reaches zero, the player is reset.
musicCntLow = 72
musicCntHi = 14

*=$0810
        sei
        ;set data direction register to allow bank switching.
        lda #$97
        sta $dd02
        lda #$3e
        sta $dd00       ;Switch to bank 1.
        lda #232
        sta $d018       ;Screen at $7800 and sprite image pointers at $7bf8-$7bff, character rom at $6000-$67ff, sprite data from $4000 to $5f00 in steps of $0100, so it does not interfere with screen memory.
        lda #$c3        ;to center the screen.
        sta $d016       
        lda #$17
        sta $d011
        lda #$00
        sta $d021
        lda #0        ;Start with sprites in front of background.
        sta $d01b
        jsr fillChars   ;Put petscii image on screen.
        jsr fillColors
        jsr copySpriteData      ;Copy sprite data to bank 1. (Initially stored in bank 0 to keep the .prg file small.)
        jsr initRnd:    ; Initialize Random Number Generator (RNG).
        jsr initSprites  ;Initialize sprites.
        lda #numMoves   ;Initialize movementCounter.
        sta movementCounter
        lda #numReps
        sta repCounter
        lda #$00
        sta repeat
        
rand:
        jsr drawRnd:      ;Draw a random number.
        lda rnd1
        sta $d020
        jsr prepareStack        ;Puts random numbers on the stack.
        ;fill character ROM.
        lda #$31
        sta $01
        ldx #$00
@loop:  lda $d000,x
        sta $6000,x
        lda $d100,x
        sta $6100,x
        lda $d200,x
        sta $6200,x
        lda $d300,x
        sta $6300,x
        lda $d400,x
        sta $6400,x
        lda $d500,x
        sta $6500,x
        lda $d600,x
        sta $6600,x
        lda $d700,x
        sta $6700,x
        inx
        bne @loop:
        
        lda #musicCntLow
        sta musicCounter
        lda #musicCntHi 
        sta musicCounter+1
        lda #$00
        jsr $0fff       ;Init music player.

        lda #$35
        sta $01

wait:   lda $d011
        bmi *-3
@wait:  lda #$01        ;The main routine is run in the borders.
        cmp $d012       
        beq @update:-4
        
        lda #$fa        
        cmp $d012
        bne @wait:
;Play music. Update sprites, which is done again in top border, so it doubles sprite movement speed.      
        ldx #safeSP
        txs
        lda #$00
        jsr $1003       ;Play music.
        dec musicCounter        ;Count down, and re-initialize player after song ends.
        bne @update:-4
        dec musicCounter+1      
        bne @update:-4
        lda #musicCntLow
        sta musicCounter
        lda #musicCntHi 
        sta musicCounter+1
        lda #$00
        jsr $0fff       ;Init music player.
        
        ldx #$00
        stx xmsb

        
@update:        ;Update sprite positions.
        clc
        inc spriteDirections,x
        ldy spriteDirections,x
        lda spriteXfrac,x
        adc XdirectionTableFrac,y
        sta spriteXfrac,x

        lda XdirectionTableLow,y
        bmi @negative:
        adc spriteXlow,x
        sta spriteXlow,x
        lda spriteXhigh,x
        adc #$00
        sta spriteXhigh,x
        jmp @setmsb:
@negative:
        adc spriteXlow,x
        sta spriteXlow,x
        lda spriteXhigh,x
        adc #$ff
        sta spriteXhigh,x
@setmsb:
        lsr
        ror xmsb
        lda spriteYfrac,x
        clc
        adc YdirectionTableFrac,y
        sta spriteYfrac,x
        lda spriteYlow,x
        adc YdirectionTableLow,y
        sta spriteYlow,x
        
@next1: inx
        cpx #$08
        bne @update:

        lda xmsb
        sta $d010
;Copy table values to vic registers.
        ldy #$07
        ldx #$0e
@copy:  lda spriteXlow,y
        sta $d000,x
        lda spriteYlow,y
        sta $d001,x
        dex
        dex
        dey
        bpl @copy:
        iny        
        sty $d015 ;Turn off sprites.
        
   
testRepeat:   
        ;Top or bottom border?
        ; Sprites are updated in top border again, so we can skip display code.
        lda $d011
        bmi decrease:
        lda repeat
        bne @tascode:
        lda spriteXlow,y
        ldx spriteYlow,y
;----------------------
;This section calculates the color ram address under the sprite.
;A contains sprite X pos, X sprite Y pos. Y contains spritenumber.
        lsr
        lsr
        lsr
        sta temp
        txa
        sec
        sbc #$28      
        lsr
        lsr
        lsr
        tax
        clc
        lda lowbyte,x
        sta selfmod:+1
        lda hibyte,x
        adc #$d8
        sta selfmod:+2    
        lda $d010
        and bit2value,y
        beq @testleft:
        lda temp
        adc #30
        jmp @store:
@testleft:
        lda temp
        sec
        sbc #02
@store:
        clc
        adc selfmod:+1
        sta selfmod:+1
        bcc *+5
        inc selfmod:+2
;------------------------------------------
@tascode:
        lax movementCounter ;Here a countervalue is used, but the registers could also be made to depend in some complicated way on sprite coordinates. For example, even on a Z coordinate if you want to do 3D stuff.
        tas $7bf8,y     ;TAS changes sprite image according to the counter value, and SP as well.
        pla
        sta $d027,y     ;Change sprite color as well.
        ldx repeat
        bne nextsprite:
selfmod:
        sta $d800       ;Adress is modified above to the value underneath the sprite.
        
nextsprite:
        iny
        cpy #$08
        bne testRepeat: 
        lda #$ff        ;Turn on sprites.
        sta $d015
decrease:        
        dec movementCounter
        bne @jumpwait:
        lda #numMoves
        sta movementCounter
        lda $d01b
        eor #$ff
        sta $d01b
        lda repeat
        eor #$01
        sta repeat
        
        bne @jumpwait:

        jsr initSprites        
        dec repCounter
        bne @jumpwait:
        ldx #safeSP     ;Change SP to avoid that the return address of prepareStack is overwritten with random numbers.
        txs
        jsr prepareStack        ;Put new random numbers on the stack.
        jsr fillColors  ;Restore petscii image colors.
        jsr drawRnd:      ;Draw a random number.
        lda rnd1
        sta $d020
        lda #numReps
        sta repCounter

@jumpwait:
        jmp wait:


        

bit2value
 BYTE 1,2,4,8,16,32,64,128
lowbyte ;Address to be added to screen memory base address.
 BYTE 0,40,80,120,160,200,240,24,64,104,144,184,224,8,48,88,128,168,208,248,32,72,112,152,192
hibyte
 BYTE 0, 0, 0,  0,  0,  0,  0, 1, 1,  1,  1,  1,  1,2, 2, 2,  2,  2,  2,  2, 3, 3,  3,  3,  3       

replace = 45    ;Replace empty space on screen with this character.

fillChars
        ldx #$00
        
@loop:  lda petsciiChars,x
        cmp #$20
        bne *+4
        lda #replace
        sta $7800,x
        lda petsciiChars+$100,x
        cmp #$20
        bne *+4
        lda #replace
        sta $7900,x
        lda petsciiChars+$200,x
        cmp #$20
        bne *+4
        lda #replace
        sta $7a00,x
        lda petsciiChars+$300,x
        cmp #$20
        bne *+4
        lda #replace
        sta $7b00,x
        inx
        bne @loop:
        rts

fillColors
        ldx #$00
@loop:
        lda petsciiColors,x
        sta $d800,x
        lda petsciiColors+$100,x
        sta $d900,x
        lda petsciiColors+$200,x
        sta $da00,x
        lda petsciiColors+$300,x
        sta $db00,x
        inx
        bne @loop:
        rts

initSprites   ;Intialize sprites.
        lda #$00
        sta $7bf8
        sta $7bf9
        sta $7bfa
        sta $7bfb
        sta $7bfc
        sta $7bfd
        sta $7bfe
        sta $7bff
        sta $d010
;Initialize movement directions...
        ldx #$07
@loop2: jsr drawRnd:
        lda rnd1
        sta spriteDirections,x
        dex
        bpl @loop2:
; and sprite initial positions.
        ldx #$07
        ldy #$0e
@loop3: lda #$00
        sta spriteXfrac,x
        sta spriteXhigh,x
        sta spriteYfrac,x
        lda #$ac
        sta spriteXlow,x
        sta $d000,y
        lda #$0e
        sta $d027,x
        lda #$8c
        sta spriteYlow,x
        sta $d001,y
        dey
        dey
        dex
        bpl @loop3:
        lda #$06
        sta $d025
        lda #$01
        sta $d026
        lda #$ff
        sta $d01c
        sta $d015
        rts

prepareStack
        tsx
        stx temp
        ldx movementCounter
        txs        
@loop   jsr drawRnd:      ;Draw random number 1-15.
        lda rnd1
        and #$0f
        beq @loop
        pha     ;Push more times, since we don't want to change char color every pixel.
        pha
        pha
        pha
        pha
        pha
        pha
        pha
        tsx
        bpl @loop       
        cpx #$f0        ;bpl only doesn't work when movementCounter > $7f.
        bcc @loop
        ldx temp        ;Restore SP.
        txs
        rts

drawRandomNumber 
initRnd:
        lda #$ff        ;SID RNG for seeding.
        sta $d40e
        sta $d40f
        lda #$80
        sta $d412
        lda $$d41b      ;seed1
        sta rnd1
        lda $dc04       ;seed2
        sta rnd2 
 
drawRnd:               ; 15 bit MLS.
        clc
        ldy #$18       ;opcode clc
        lda #$60
        rol rnd1
        rla rnd2
        beq @carry:
        eor #$60
        cmp #$01
        bcc @carry:
        ldy #$38        ;opcocde sec
@carry: sty drawRnd:

        rts

spriteDirections
;Sprites can move in 256 different radial directions (table is filled with random numbers).
 BYTE 1,1,1,1,1,1,1,1

spriteXfrac
;We calculate sprite X positions with one fractional byte.
 BYTE 0,0,0,0,0,0,0,0
spriteXlow
;LSB of sprite X positions.
 BYTE 172,172,172,172,172,172,172,172
spriteXhigh
;MSB
 BYTE 0,0,0,0,0,0,0,0

spriteYfrac
;We calculate sprite Y positions with one fractional byte.
 BYTE 0,0,0,0,0,0,0,0
spriteYlow
;LSB of sprite Y positions.
 BYTE 172,172,172,172,172,172,172,172

*=$0b80
;Delta X and Y added to the sprite positions (uses one fractional byte).
XdirectionTableFrac
incBin "Xfrac.bin"

XdirectionTableLow
incBin "Xlow.bin"

YdirectionTableFrac
incBin "Yfrac.bin"

YdirectionTableLow
incBin "Ylow.bin"

*=$0fff
;Include music file "Ambient.sid" (http://csdb.dk/sid/?id=2658), by 6R6 (Glenn Rune Gallefoss).
incBin "Ambient.sid", $7e

*=$1c00
;Sprites are taken from the "Subchrist - Examples 1" file that comes with SpritePad V2.0 Beta 1 (http://csdb.dk/release/?id=132081), by Stewart Wilson.
incBin "Sprites.bin"

*=$2400
;Colors of the petscii image "Star Power", by Endurion (http://csdb.dk/release/?id=122592).
petsciiColors
incBin "StarPowerColors.bin"

*=$2800
;Characters of that petscii image.
petsciiChars
incBin "StarPowerChars.bin"


copySpriteData
;Copies sprite data from bank 0 to the required locations in bank 1.
        ldx #$3f
        lda $1c00,x
        sta $4000,x
        lda $1c40,x
        sta $4100,x
        lda $1c80,x
        sta $4200,x
        lda $1cc0,x
        sta $4300,x
        lda $1d00,x
        sta $4400,x
        lda $1d40,x
        sta $4500,x
        lda $1d80,x
        sta $4600,x
        lda $1dc0,x
        sta $4700,x
        lda $1e00,x
        sta $4800,x
        lda $1e40,x
        sta $4900,x
        lda $1e80,x
        sta $4a00,x
        lda $1ec0,x
        sta $4b00,x
        lda $1f00,x
        sta $4c00,x
        lda $1f40,x
        sta $4d00,x
        lda $1f80,x
        sta $4e00,x
        lda $1fc0,x
        sta $4f00,x

        lda $2000,x
        sta $5000,x
        lda $2040,x
        sta $5100,x
        lda $2080,x
        sta $5200,x
        lda $20c0,x
        sta $5300,x
        lda $2100,x
        sta $5400,x
        lda $2140,x
        sta $5500,x
        lda $2180,x
        sta $5600,x
        lda $21c0,x
        sta $5700,x
        lda $2200,x
        sta $5800,x
        lda $2240,x
        sta $5900,x
        lda $2280,x
        sta $5a00,x
        lda $22c0,x
        sta $5b00,x
        lda $2300,x
        sta $5c00,x
        lda $2340,x
        sta $5d00,x
        lda $2380,x
        sta $5e00,x
        lda $23c0,x
        sta $5f00,x
        
        dex
        bmi *+5
        jmp copySpriteData+2
        rts

